# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

from typing import Set, Tuple, Union

from libcst import Import, ImportFrom, ImportStar, Module
from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand
from libcst.codemod.visitors import GatherCommentsVisitor, RemoveImportsVisitor
from libcst.helpers import get_absolute_module_for_import
from libcst.metadata import PositionProvider, ProviderT


DEFAULT_SUPPRESS_COMMENT_REGEX = (
    r".*\W(noqa|lint-ignore: ?unused-import|lint-ignore: ?F401)(\W.*)?$"
)


class RemoveUnusedImportsCommand(VisitorBasedCodemodCommand):
    """
    Remove all unused imports from a file based on scope analysis.

    This command analyses individual files in isolation and does not attempt
    to track cross-references between them. If a symbol is imported in a file
    but otherwise unused in it, that import will be removed even if it is being
    referenced from another file.
    """

    DESCRIPTION: str = (
        "Remove all imports that are not used in a file. "
        + "Note: only considers the file in isolation. "
    )

    METADATA_DEPENDENCIES: Tuple[ProviderT] = (PositionProvider,)

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)
        self._ignored_lines: Set[int] = set()

    def visit_Module(self, node: Module) -> bool:
        comment_visitor = GatherCommentsVisitor(
            self.context, DEFAULT_SUPPRESS_COMMENT_REGEX
        )
        node.visit(comment_visitor)
        self._ignored_lines = set(comment_visitor.comments.keys())
        return True

    def visit_Import(self, node: Import) -> bool:
        self._handle_import(node)
        return False

    def visit_ImportFrom(self, node: ImportFrom) -> bool:
        self._handle_import(node)
        return False

    def _handle_import(self, node: Union[Import, ImportFrom]) -> None:
        node_start = self.get_metadata(PositionProvider, node).start.line
        if node_start in self._ignored_lines:
            return

        names = node.names
        if isinstance(names, ImportStar):
            return

        for alias in names:
            position = self.get_metadata(PositionProvider, alias)
            lines = set(range(position.start.line, position.end.line + 1))
            if lines.isdisjoint(self._ignored_lines):
                if isinstance(node, Import):
                    RemoveImportsVisitor.remove_unused_import(
                        self.context,
                        module=alias.evaluated_name,
                        asname=alias.evaluated_alias,
                    )
                else:
                    module_name = get_absolute_module_for_import(
                        self.context.full_module_name, node
                    )
                    if module_name is None:
                        raise ValueError(
                            f"Couldn't get absolute module name for {alias.evaluated_name}"
                        )
                    RemoveImportsVisitor.remove_unused_import(
                        self.context,
                        module=module_name,
                        obj=alias.evaluated_name,
                        asname=alias.evaluated_alias,
                    )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class Production:
    name: str
    children: str
    version: Optional[str]
    future: Optional[str]

    def __str__(self) -> str:
        return f"{self.name}: {self.children}"



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from typing import Any

import libcst as cst
from libcst import parse_expression
from libcst._nodes.tests.base import CSTNodeTest, parse_expression_as
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


def _parse_expression_force_38(code: str) -> cst.BaseExpression:
    return cst.parse_expression(
        code, config=cst.PartialParserConfig(python_version="3.8")
    )


class AtomTest(CSTNodeTest):
    @data_provider(
        (
            # Simple identifier
            {
                "node": cst.Name("test"),
                "code": "test",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Parenthesized identifier
            {
                "node": cst.Name(
                    "test", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                ),
                "code": "(test)",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 1), (1, 5)),
            },
            # Decimal integers
            {
                "node": cst.Integer("12345"),
                "code": "12345",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Integer("0000"),
                "code": "0000",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Integer("1_234_567"),
                "code": "1_234_567",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Integer("0_000"),
                "code": "0_000",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Binary integers
            {
                "node": cst.Integer("0b0000"),
                "code": "0b0000",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Integer("0B1011_0100"),
                "code": "0B1011_0100",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Octal integers
            {
                "node": cst.Integer("0o12345"),
                "code": "0o12345",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Integer("0O12_345"),
                "code": "0O12_345",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Hex numbers
            {
                "node": cst.Integer("0x123abc"),
                "code": "0x123abc",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Integer("0X12_3ABC"),
                "code": "0X12_3ABC",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Parenthesized integers
            {
                "node": cst.Integer(
                    "123", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                ),
                "code": "(123)",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 1), (1, 4)),
            },
            # Non-exponent floats
            {
                "node": cst.Float("12345."),
                "code": "12345.",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Float("00.00"),
                "code": "00.00",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Float("12.21"),
                "code": "12.21",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Float(".321"),
                "code": ".321",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Float("1_234_567."),
                "code": "1_234_567.",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Float("0.000_000"),
                "code": "0.000_000",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Exponent floats
            {
                "node": cst.Float("12345.e10"),
                "code": "12345.e10",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Float("00.00e10"),
                "code": "00.00e10",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Float("12.21e10"),
                "code": "12.21e10",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Float(".321e10"),
                "code": ".321e10",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Float("1_234_567.e10"),
                "code": "1_234_567.e10",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Float("0.000_000e10"),
                "code": "0.000_000e10",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Float("1e+10"),
                "code": "1e+10",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Float("1e-10"),
                "code": "1e-10",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Parenthesized floats
            {
                "node": cst.Float(
                    "123.4", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                ),
                "code": "(123.4)",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 1), (1, 6)),
            },
            # Imaginary numbers
            {
                "node": cst.Imaginary("12345j"),
                "code": "12345j",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Imaginary("1_234_567J"),
                "code": "1_234_567J",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Imaginary("12345.e10j"),
                "code": "12345.e10j",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Imaginary(".321J"),
                "code": ".321J",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Parenthesized imaginary
            {
                "node": cst.Imaginary(
                    "123.4j", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                ),
                "code": "(123.4j)",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 1), (1, 7)),
            },
            # Simple elipses
            {
                "node": cst.Ellipsis(),
                "code": "...",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Parenthesized elipses
            {
                "node": cst.Ellipsis(lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)),
                "code": "(...)",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 1), (1, 4)),
            },
            # Simple strings
            {
                "node": cst.SimpleString('""'),
                "code": '""',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.SimpleString("''"),
                "code": "''",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.SimpleString('"test"'),
                "code": '"test"',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.SimpleString('b"test"'),
                "code": 'b"test"',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.SimpleString('r"test"'),
                "code": 'r"test"',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.SimpleString('"""test"""'),
                "code": '"""test"""',
                "parser": parse_expression,
                "expected_position": None,
            },
            # Validate parens
            {
                "node": cst.SimpleString(
                    '"test"', lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                ),
                "code": '("test")',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.SimpleString(
                    'rb"test"', lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                ),
                "code": '(rb"test")',
                "parser": parse_expression,
                "expected_position": CodeRange((1, 1), (1, 9)),
            },
            # Test that _safe_to_use_with_word_operator allows no space around quotes
            {
                "node": cst.Comparison(
                    cst.SimpleString('"a"'),
                    [
                        cst.ComparisonTarget(
                            cst.In(
                                whitespace_before=cst.SimpleWhitespace(""),
                                whitespace_after=cst.SimpleWhitespace(""),
                            ),
                            cst.SimpleString('"abc"'),
                        )
                    ],
                ),
                "code": '"a"in"abc"',
                "parser": parse_expression,
            },
            {
                "node": cst.Comparison(
                    cst.SimpleString('"a"'),
                    [
                        cst.ComparisonTarget(
                            cst.In(
                                whitespace_before=cst.SimpleWhitespace(""),
                                whitespace_after=cst.SimpleWhitespace(""),
                            ),
                            cst.ConcatenatedString(
                                cst.SimpleString('"a"'), cst.SimpleString('"bc"')
                            ),
                        )
                    ],
                ),
                "code": '"a"in"a""bc"',
                "parser": parse_expression,
            },
            # Parenthesis make no spaces around a prefix okay
            {
                "node": cst.Comparison(
                    cst.SimpleString('b"a"'),
                    [
                        cst.ComparisonTarget(
                            cst.In(
                                whitespace_before=cst.SimpleWhitespace(""),
                                whitespace_after=cst.SimpleWhitespace(""),
                            ),
                            cst.SimpleString(
                                'b"abc"',
                                lpar=[cst.LeftParen()],
                                rpar=[cst.RightParen()],
                            ),
                        )
                    ],
                ),
                "code": 'b"a"in(b"abc")',
                "parser": parse_expression,
            },
            {
                "node": cst.Comparison(
                    cst.SimpleString('b"a"'),
                    [
                        cst.ComparisonTarget(
                            cst.In(
                                whitespace_before=cst.SimpleWhitespace(""),
                                whitespace_after=cst.SimpleWhitespace(""),
                            ),
                            cst.ConcatenatedString(
                                cst.SimpleString('b"a"'),
                                cst.SimpleString('b"bc"'),
                                lpar=[cst.LeftParen()],
                                rpar=[cst.RightParen()],
                            ),
                        )
                    ],
                ),
                "code": 'b"a"in(b"a"b"bc")',
                "parser": parse_expression,
            },
            # Empty formatted strings
            {
                "node": cst.FormattedString(start='f"', parts=(), end='"'),
                "code": 'f""',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(start="f'", parts=(), end="'"),
                "code": "f''",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(start='f"""', parts=(), end='"""'),
                "code": 'f""""""',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(start="f'''", parts=(), end="'''"),
                "code": "f''''''",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Non-empty formatted strings
            {
                "node": cst.FormattedString(parts=(cst.FormattedStringText("foo"),)),
                "code": 'f"foo"',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(
                    parts=(cst.FormattedStringExpression(cst.Name("foo")),)
                ),
                "code": 'f"{foo}"',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringText("foo "),
                        cst.FormattedStringExpression(cst.Name("bar")),
                        cst.FormattedStringText(" baz"),
                    )
                ),
                "code": 'f"foo {bar} baz"',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringText("foo "),
                        cst.FormattedStringExpression(cst.Call(cst.Name("bar"))),
                        cst.FormattedStringText(" baz"),
                    )
                ),
                "code": 'f"foo {bar()} baz"',
                "parser": parse_expression,
                "expected_position": None,
            },
            # Formatted strings with conversions and format specifiers
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringExpression(cst.Name("foo"), conversion="s"),
                    )
                ),
                "code": 'f"{foo!s}"',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringExpression(cst.Name("foo"), format_spec=()),
                    )
                ),
                "code": 'f"{foo:}"',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringExpression(
                            cst.Name("today"),
                            format_spec=(cst.FormattedStringText("%B %d, %Y"),),
                        ),
                    )
                ),
                "code": 'f"{today:%B %d, %Y}"',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringExpression(
                            cst.Name("foo"),
                            format_spec=(
                                cst.FormattedStringExpression(cst.Name("bar")),
                            ),
                        ),
                    )
                ),
                "code": 'f"{foo:{bar}}"',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringExpression(
                            cst.Name("foo"),
                            format_spec=(
                                cst.FormattedStringExpression(cst.Name("bar")),
                                cst.FormattedStringText("."),
                                cst.FormattedStringExpression(cst.Name("baz")),
                            ),
                        ),
                    )
                ),
                "code": 'f"{foo:{bar}.{baz}}"',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringExpression(
                            cst.Name("foo"),
                            conversion="s",
                            format_spec=(
                                cst.FormattedStringExpression(cst.Name("bar")),
                            ),
                        ),
                    )
                ),
                "code": 'f"{foo!s:{bar}}"',
                "parser": parse_expression,
                "expected_position": None,
            },
            # Test equality expression added in 3.8.
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringExpression(
                            cst.Name("foo"),
                            equal=cst.AssignEqual(
                                whitespace_before=cst.SimpleWhitespace(""),
                                whitespace_after=cst.SimpleWhitespace(""),
                            ),
                        ),
                    ),
                ),
                "code": 'f"{foo=}"',
                "parser": _parse_expression_force_38,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringExpression(
                            cst.Name("foo"),
                            equal=cst.AssignEqual(
                                whitespace_before=cst.SimpleWhitespace(""),
                                whitespace_after=cst.SimpleWhitespace(""),
                            ),
                            conversion="s",
                        ),
                    ),
                ),
                "code": 'f"{foo=!s}"',
                "parser": _parse_expression_force_38,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringExpression(
                            cst.Name("foo"),
                            equal=cst.AssignEqual(
                                whitespace_before=cst.SimpleWhitespace(""),
                                whitespace_after=cst.SimpleWhitespace(""),
                            ),
                            conversion="s",
                            format_spec=(
                                cst.FormattedStringExpression(cst.Name("bar")),
                            ),
                        ),
                    ),
                ),
                "code": 'f"{foo=!s:{bar}}"',
                "parser": _parse_expression_force_38,
                "expected_position": None,
            },
            # Test that equality support doesn't break existing support
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringExpression(
                            cst.Comparison(
                                left=cst.Name(
                                    value="a",
                                ),
                                comparisons=[
                                    cst.ComparisonTarget(
                                        operator=cst.Equal(),
                                        comparator=cst.Name(
                                            value="b",
                                        ),
                                    ),
                                ],
                            ),
                        ),
                    ),
                ),
                "code": 'f"{a == b}"',
                "parser": _parse_expression_force_38,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringExpression(
                            cst.Comparison(
                                left=cst.Name(
                                    value="a",
                                ),
                                comparisons=[
                                    cst.ComparisonTarget(
                                        operator=cst.NotEqual(),
                                        comparator=cst.Name(
                                            value="b",
                                        ),
                                    ),
                                ],
                            ),
                        ),
                    ),
                ),
                "code": 'f"{a != b}"',
                "parser": _parse_expression_force_38,
                "expected_position": None,
            },
            {
                "node": cst.FormattedString(
                    parts=(
                        cst.FormattedStringExpression(
                            cst.NamedExpr(
                                target=cst.Name(
                                    value="a",
                                ),
                                value=cst.Integer(
                                    value="5",
                                ),
                                lpar=(cst.LeftParen(),),
                                rpar=(cst.RightParen(),),
                            ),
                        ),
                    ),
                ),
                "code": 'f"{(a := 5)}"',
                "parser": _parse_expression_force_38,
                "expected_position": None,
            },
            # Validate parens
            {
                "node": cst.FormattedString(
                    start='f"',
                    parts=(),
                    end='"',
                    lpar=(cst.LeftParen(),),
                    rpar=(cst.RightParen(),),
                ),
                "code": '(f"")',
                "parser": parse_expression,
                "expected_position": CodeRange((1, 1), (1, 4)),
            },
            # Generator expression (doesn't make sense, but legal syntax)
            {
                "node": cst.FormattedString(
                    start='f"',
                    parts=[
                        cst.FormattedStringExpression(
                            expression=cst.GeneratorExp(
                                elt=cst.Name(
                                    value="x",
                                ),
                                for_in=cst.CompFor(
                                    target=cst.Name(
                                        value="x",
                                    ),
                                    iter=cst.Name(
                                        value="y",
                                    ),
                                ),
                                lpar=[],
                                rpar=[],
                            ),
                        ),
                    ],
                    end='"',
                ),
                "code": 'f"{x for x in y}"',
                "parser": parse_expression,
                "expected_position": None,
            },
            # Concatenated strings
            {
                "node": cst.ConcatenatedString(
                    cst.SimpleString('"ab"'), cst.SimpleString('"c"')
                ),
                "code": '"ab""c"',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.ConcatenatedString(
                    cst.SimpleString('"ab"'),
                    cst.ConcatenatedString(
                        cst.SimpleString('"c"'), cst.SimpleString('"d"')
                    ),
                ),
                "code": '"ab""c""d"',
                "parser": parse_expression,
                "expected_position": None,
            },
            # mixed SimpleString and FormattedString
            {
                "node": cst.ConcatenatedString(
                    cst.FormattedString([cst.FormattedStringText("ab")]),
                    cst.SimpleString('"c"'),
                ),
                "code": 'f"ab""c"',
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.ConcatenatedString(
                    cst.SimpleString('"ab"'),
                    cst.FormattedString([cst.FormattedStringText("c")]),
                ),
                "code": '"ab"f"c"',
                "parser": parse_expression,
                "expected_position": None,
            },
            # Concatenated parenthesized strings
            {
                "node": cst.ConcatenatedString(
                    lpar=(cst.LeftParen(),),
                    left=cst.SimpleString('"ab"'),
                    right=cst.SimpleString('"c"'),
                    rpar=(cst.RightParen(),),
                ),
                "code": '("ab""c")',
                "parser": parse_expression,
                "expected_position": None,
            },
            # Validate spacing
            {
                "node": cst.ConcatenatedString(
                    lpar=(cst.LeftParen(whitespace_after=cst.SimpleWhitespace(" ")),),
                    left=cst.SimpleString('"ab"'),
                    whitespace_between=cst.SimpleWhitespace(" "),
                    right=cst.SimpleString('"c"'),
                    rpar=(cst.RightParen(whitespace_before=cst.SimpleWhitespace(" ")),),
                ),
                "code": '( "ab" "c" )',
                "parser": parse_expression,
                "expected_position": CodeRange((1, 2), (1, 10)),
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        # We don't have sentinel nodes for atoms, so we know that 100% of atoms
        # can be parsed identically to their creation.
        self.validate_node(**kwargs)

    @data_provider(
        (
            {
                "node": cst.FormattedStringExpression(
                    cst.Name("today"),
                    format_spec=(cst.FormattedStringText("%B %d, %Y"),),
                ),
                "code": "{today:%B %d, %Y}",
                "parser": None,
                "expected_position": CodeRange((1, 0), (1, 17)),
            },
        )
    )
    def test_valid_no_parse(self, **kwargs: Any) -> None:
        # Test some nodes that aren't valid source code by themselves
        self.validate_node(**kwargs)

    @data_provider(
        (
            # Expression wrapping parenthesis rules
            {
                "get_node": (lambda: cst.Name("foo", lpar=(cst.LeftParen(),))),
                "expected_re": "left paren without right paren",
            },
            {
                "get_node": lambda: cst.Name("foo", rpar=(cst.RightParen(),)),
                "expected_re": "right paren without left paren",
            },
            {
                "get_node": lambda: cst.Ellipsis(lpar=(cst.LeftParen(),)),
                "expected_re": "left paren without right paren",
            },
            {
                "get_node": lambda: cst.Ellipsis(rpar=(cst.RightParen(),)),
                "expected_re": "right paren without left paren",
            },
            {
                "get_node": lambda: cst.Integer("5", lpar=(cst.LeftParen(),)),
                "expected_re": "left paren without right paren",
            },
            {
                "get_node": lambda: cst.Integer("5", rpar=(cst.RightParen(),)),
                "expected_re": "right paren without left paren",
            },
            {
                "get_node": lambda: cst.Float("5.5", lpar=(cst.LeftParen(),)),
                "expected_re": "left paren without right paren",
            },
            {
                "get_node": lambda: cst.Float("5.5", rpar=(cst.RightParen(),)),
                "expected_re": "right paren without left paren",
            },
            {
                "get_node": (lambda: cst.Imaginary("5j", lpar=(cst.LeftParen(),))),
                "expected_re": "left paren without right paren",
            },
            {
                "get_node": (lambda: cst.Imaginary("5j", rpar=(cst.RightParen(),))),
                "expected_re": "right paren without left paren",
            },
            {
                "get_node": (lambda: cst.Integer("5", lpar=(cst.LeftParen(),))),
                "expected_re": "left paren without right paren",
            },
            {
                "get_node": (lambda: cst.Integer("5", rpar=(cst.RightParen(),))),
                "expected_re": "right paren without left paren",
            },
            {
                "get_node": (
                    lambda: cst.SimpleString("'foo'", lpar=(cst.LeftParen(),))
                ),
                "expected_re": "left paren without right paren",
            },
            {
                "get_node": (
                    lambda: cst.SimpleString("'foo'", rpar=(cst.RightParen(),))
                ),
                "expected_re": "right paren without left paren",
            },
            {
                "get_node": (
                    lambda: cst.FormattedString(parts=(), lpar=(cst.LeftParen(),))
                ),
                "expected_re": "left paren without right paren",
            },
            {
                "get_node": (
                    lambda: cst.FormattedString(parts=(), rpar=(cst.RightParen(),))
                ),
                "expected_re": "right paren without left paren",
            },
            {
                "get_node": (
                    lambda: cst.ConcatenatedString(
                        cst.SimpleString("'foo'"),
                        cst.SimpleString("'foo'"),
                        lpar=(cst.LeftParen(),),
                    )
                ),
                "expected_re": "left paren without right paren",
            },
            {
                "get_node": (
                    lambda: cst.ConcatenatedString(
                        cst.SimpleString("'foo'"),
                        cst.SimpleString("'foo'"),
                        rpar=(cst.RightParen(),),
                    )
                ),
                "expected_re": "right paren without left paren",
            },
            # Node-specific rules
            {
                "get_node": (lambda: cst.Name("")),
                "expected_re": "empty name identifier",
            },
            {
                "get_node": (lambda: cst.Name(r"\/")),
                "expected_re": "not a valid identifier",
            },
            {
                "get_node": (lambda: cst.Integer("")),
                "expected_re": "not a valid integer",
            },
            {
                "get_node": (lambda: cst.Integer("012345")),
                "expected_re": "not a valid integer",
            },
            {
                "get_node": (lambda: cst.Integer("012345")),
                "expected_re": "not a valid integer",
            },
            {
                "get_node": (lambda: cst.Integer("_12345")),
                "expected_re": "not a valid integer",
            },
            {
                "get_node": (lambda: cst.Integer("0b2")),
                "expected_re": "not a valid integer",
            },
            {
                "get_node": (lambda: cst.Integer("0o8")),
                "expected_re": "not a valid integer",
            },
            {
                "get_node": (lambda: cst.Integer("0xg")),
                "expected_re": "not a valid integer",
            },
            {
                "get_node": (lambda: cst.Integer("123.45")),
                "expected_re": "not a valid integer",
            },
            {
                "get_node": (lambda: cst.Integer("12345j")),
                "expected_re": "not a valid integer",
            },
            {
                "get_node": (lambda: cst.Float("12.3.45")),
                "expected_re": "not a valid float",
            },
            {"get_node": (lambda: cst.Float("12")), "expected_re": "not a valid float"},
            {
                "get_node": (lambda: cst.Float("12.3j")),
                "expected_re": "not a valid float",
            },
            {
                "get_node": (lambda: cst.Imaginary("_12345j")),
                "expected_re": "not a valid imaginary",
            },
            {
                "get_node": (lambda: cst.Imaginary("0b0j")),
                "expected_re": "not a valid imaginary",
            },
            {
                "get_node": (lambda: cst.Imaginary("0o0j")),
                "expected_re": "not a valid imaginary",
            },
            {
                "get_node": (lambda: cst.Imaginary("0x0j")),
                "expected_re": "not a valid imaginary",
            },
            {
                "get_node": (lambda: cst.SimpleString('wee""')),
                "expected_re": "Invalid string prefix",
            },
            {
                "get_node": (lambda: cst.SimpleString("'")),
                "expected_re": "must have enclosing quotes",
            },
            {
                "get_node": (lambda: cst.SimpleString('"')),
                "expected_re": "must have enclosing quotes",
            },
            {
                "get_node": (lambda: cst.SimpleString("\"'")),
                "expected_re": "must have matching enclosing quotes",
            },
            {
                "get_node": (lambda: cst.SimpleString("")),
                "expected_re": "must have enclosing quotes",
            },
            {
                "get_node": (lambda: cst.SimpleString("'bla")),
                "expected_re": "must have matching enclosing quotes",
            },
            {
                "get_node": (lambda: cst.SimpleString("f''")),
                "expected_re": "Invalid string prefix",
            },
            {
                "get_node": (lambda: cst.SimpleString("'''bla''")),
                "expected_re": "must have matching enclosing quotes",
            },
            {
                "get_node": (lambda: cst.SimpleString("'''bla\"\"\"")),
                "expected_re": "must have matching enclosing quotes",
            },
            {
                "get_node": (lambda: cst.FormattedString(start="'", parts=(), end="'")),
                "expected_re": "Invalid f-string prefix",
            },
            {
                "get_node": (
                    lambda: cst.FormattedString(start="f'", parts=(), end='"')
                ),
                "expected_re": "must have matching enclosing quotes",
            },
            {
                "get_node": (
                    lambda: cst.ConcatenatedString(
                        cst.SimpleString(
                            '"ab"', lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                        ),
                        cst.SimpleString('"c"'),
                    )
                ),
                "expected_re": "Cannot concatenate parenthesized",
            },
            {
                "get_node": (
                    lambda: cst.ConcatenatedString(
                        cst.SimpleString('"ab"'),
                        cst.SimpleString(
                            '"c"', lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                        ),
                    )
                ),
                "expected_re": "Cannot concatenate parenthesized",
            },
            {
                "get_node": (
                    lambda: cst.ConcatenatedString(
                        cst.SimpleString('"ab"'), cst.SimpleString('b"c"')
                    )
                ),
                "expected_re": "Cannot concatenate string and bytes",
            },
            # This isn't valid code: `"a" inb"abc"`
            {
                "get_node": (
                    lambda: cst.Comparison(
                        cst.SimpleString('"a"'),
                        [
                            cst.ComparisonTarget(
                                cst.In(whitespace_after=cst.SimpleWhitespace("")),
                                cst.SimpleString('b"abc"'),
                            )
                        ],
                    )
                ),
                "expected_re": "Must have at least one space around comparison operator.",
            },
            # Also not valid: `"a" in b"a"b"bc"`
            {
                "get_node": (
                    lambda: cst.Comparison(
                        cst.SimpleString('"a"'),
                        [
                            cst.ComparisonTarget(
                                cst.In(whitespace_after=cst.SimpleWhitespace("")),
                                cst.ConcatenatedString(
                                    cst.SimpleString('b"a"'), cst.SimpleString('b"bc"')
                                ),
                            )
                        ],
                    )
                ),
                "expected_re": "Must have at least one space around comparison operator.",
            },
        )
    )
    def test_invalid(self, **kwargs: Any) -> None:
        self.assert_invalid(**kwargs)

    @data_provider(
        (
            {
                "code": "u'x'",
                "parser": parse_expression_as(python_version="3.3"),
                "expect_success": True,
            },
            {
                "code": "u'x'",
                "parser": parse_expression_as(python_version="3.1"),
                "expect_success": False,
            },
        )
    )
    def test_versions(self, **kwargs: Any) -> None:
        self.assert_parses(**kwargs)


class StringHelperTest(CSTNodeTest):
    def test_string_prefix_and_quotes(self) -> None:
        """
        Test our helpers out for various strings.
        """
        emptybytestring = cst.ensure_type(parse_expression('b""'), cst.SimpleString)
        bytestring = cst.ensure_type(parse_expression('b"abc"'), cst.SimpleString)
        multilinestring = cst.ensure_type(parse_expression('""""""'), cst.SimpleString)
        formatstring = cst.ensure_type(parse_expression('f""""""'), cst.FormattedString)

        self.assertEqual(emptybytestring.prefix, "b")
        self.assertEqual(emptybytestring.quote, '"')
        self.assertEqual(emptybytestring.raw_value, "")
        self.assertEqual(bytestring.prefix, "b")
        self.assertEqual(bytestring.quote, '"')
        self.assertEqual(bytestring.raw_value, "abc")
        self.assertEqual(multilinestring.prefix, "")
        self.assertEqual(multilinestring.quote, '"""')
        self.assertEqual(multilinestring.raw_value, "")
        self.assertEqual(formatstring.prefix, "f")
        self.assertEqual(formatstring.quote, '"""')



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pickle
from typing import Union

import libcst as cst
import libcst.matchers as m
from libcst.matchers import (
    MatchDecoratorMismatch,
    MatcherDecoratableTransformer,
    MatcherDecoratableVisitor,
    leave,
    visit,
)
from libcst.testing.utils import UnitTest


class MatchersVisitLeaveDecoratorTypingTest(UnitTest):
    def test_valid_collector_simple(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @visit(m.SimpleString())
            def _string_visit(self, node: cst.SimpleString) -> None:
                pass

            @leave(m.SimpleString())
            def _string_leave(self, original_node: cst.SimpleString) -> None:
                pass

        # Instantiating this class should not raise any errors
        TestVisitor()

    def test_valid_transformer_simple(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @visit(m.SimpleString())
            def _string_visit(self, node: cst.SimpleString) -> None:
                pass

            @leave(m.SimpleString())
            def _string_leave(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> cst.SimpleString:
                return updated_node

        # Instantiating this class should not raise any errors
        TestVisitor()

    def test_valid_transformer_base_class(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.SimpleString())
            def _string_leave(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> cst.BaseExpression:
                return updated_node

        # Instantiating this class should not raise any errors
        TestVisitor()

    def test_valid_collector_visit_union(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @visit(m.SimpleString() | m.Name())
            def _string_visit(self, node: Union[cst.SimpleString, cst.Name]) -> None:
                pass

        # Instantiating this class should not raise any errors
        TestVisitor()

    def test_valid_transformer_visit_union(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @visit(m.SimpleString() | m.Name())
            def _string_visit(self, node: Union[cst.SimpleString, cst.Name]) -> None:
                pass

        # Instantiating this class should not raise any errors
        TestVisitor()

    def test_valid_collector_visit_superclass(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @visit(m.SimpleString() | m.Name())
            def _string_visit(self, node: cst.BaseExpression) -> None:
                pass

        # Instantiating this class should not raise any errors
        TestVisitor()

    def test_valid_transformer_visit_superclass(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @visit(m.SimpleString() | m.Name())
            def _string_visit(self, node: cst.BaseExpression) -> None:
                pass

        # Instantiating this class should not raise any errors
        TestVisitor()

    def test_valid_collector_leave_union(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @leave(m.SimpleString() | m.Name())
            def _string_leave(self, node: Union[cst.SimpleString, cst.Name]) -> None:
                pass

        # Instantiating this class should not raise any errors
        TestVisitor()

    def test_valid_transformer_leave_union(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.SimpleString() | m.Name())
            def _string_leave(
                self,
                original_node: Union[cst.SimpleString, cst.Name],
                updated_node: Union[cst.SimpleString, cst.Name],
            ) -> Union[cst.SimpleString, cst.Name]:
                return updated_node

        # Instantiating this class should not raise any errors
        TestVisitor()

    def test_valid_collector_leave_superclass(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @leave(m.SimpleString() | m.Name())
            def _string_leave(self, node: cst.BaseExpression) -> None:
                pass

        # Instantiating this class should not raise any errors
        TestVisitor()

    def test_valid_transformer_leave_superclass(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.SimpleString() | m.Name())
            def _string_leave(
                self,
                original_node: cst.BaseExpression,
                updated_node: cst.BaseExpression,
            ) -> cst.BaseExpression:
                return updated_node

        # Instantiating this class should not raise any errors
        TestVisitor()

    def test_valid_transformer_leave_return_maybe(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.AssignEqual())
            def _assign_equal_leave(
                self, original_node: cst.AssignEqual, updated_node: cst.AssignEqual
            ) -> Union[cst.AssignEqual, cst.MaybeSentinel]:
                return updated_node

        # Instantiating this class should not raise any errors
        TestVisitor()

    def test_valid_transformer_leave_return_remove(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.AssignTarget())
            def _string_visit(
                self, original_node: cst.AssignTarget, updated_node: cst.AssignTarget
            ) -> Union[cst.AssignTarget, cst.RemovalSentinel]:
                return updated_node

        # Instantiating this class should not raise any errors
        TestVisitor()

    def test_invalid_collector_visit_return(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @visit(m.SimpleString())
            def _string_visit(self, node: cst.SimpleString) -> bool:
                return False

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@visit should only decorate functions that do not return",
        ):
            TestVisitor()

    def test_invalid_transformer_visit_return(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @visit(m.SimpleString())
            def _string_visit(self, node: cst.SimpleString) -> bool:
                return False

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@visit should only decorate functions that do not return",
        ):
            TestVisitor()

    def test_invalid_transformer_visit_num_params(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @visit(m.SimpleString())
            def _string_visit(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> None:
                pass

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@visit should decorate functions which take 1 parameter",
        ):
            TestVisitor()

    def test_invalid_collector_visit_num_params(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @visit(m.SimpleString())
            def _string_visit(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> None:
                pass

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@visit should decorate functions which take 1 parameter",
        ):
            TestVisitor()

    def test_invalid_transformer_leave_num_params(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.SimpleString())
            def _string_leave(
                self, original_node: cst.SimpleString
            ) -> cst.SimpleString:
                return original_node

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave should decorate functions which take 2 parameters",
        ):
            TestVisitor()

    def test_invalid_collector_leave_num_params(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @leave(m.SimpleString())
            def _string_leave(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> None:
                pass

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave should decorate functions which take 1 parameter",
        ):
            TestVisitor()

    def test_invalid_collector_leave_return(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @leave(m.SimpleString())
            def _string_leave(self, original_node: cst.SimpleString) -> bool:
                return False

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave should only decorate functions that do not return",
        ):
            TestVisitor()

    def test_invalid_transformer_leave_return_invalid_superclass(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.SimpleString())
            def _string_visit(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> cst.BaseParenthesizableWhitespace:
                return cst.SimpleWhitespace("")

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave decorated function cannot return the type BaseParenthesizableWhitespace",
        ):
            TestVisitor()

    def test_invalid_transformer_leave_return_wrong_type(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.SimpleString())
            def _string_visit(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> cst.Pass:
                return cst.Pass()

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave decorated function cannot return the type Pass",
        ):
            TestVisitor()

    def test_invalid_transformer_leave_return_invalid_maybe(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.SimpleString())
            def _string_visit(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> Union[cst.SimpleString, cst.MaybeSentinel]:
                return updated_node

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave decorated function cannot return the type MaybeSentinel",
        ):
            TestVisitor()

    def test_invalid_transformer_leave_return_invalid_remove(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.SimpleString())
            def _string_visit(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> Union[cst.SimpleString, cst.RemovalSentinel]:
                return updated_node

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave decorated function cannot return the type RemovalSentinel",
        ):
            TestVisitor()

    def test_invalid_transformer_leave_return_invalid_union(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.SimpleString() | m.Name())
            def _string_leave(
                self,
                original_node: Union[cst.SimpleString, cst.Name],
                updated_node: Union[cst.SimpleString, cst.Name],
            ) -> Union[cst.SimpleString, cst.Pass]:
                return cst.SimpleString('""')

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave decorated function cannot return the type Pass",
        ):
            TestVisitor()

    def test_invalid_collector_visit_union(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @visit(m.SimpleString() | m.Name())
            def _string_visit(self, node: cst.SimpleString) -> None:
                pass

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@visit can be called with Name but the decorated function parameter annotations do not include this type",
        ):
            TestVisitor()

    def test_invalid_transformer_visit_union(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @visit(m.SimpleString() | m.Name())
            def _string_visit(self, node: cst.SimpleString) -> None:
                pass

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@visit can be called with Name but the decorated function parameter annotations do not include this type",
        ):
            TestVisitor()

    def test_invalid_collector_visit_superclass(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @visit(m.SimpleString() | m.Pass())
            def _string_visit(self, node: cst.BaseExpression) -> None:
                pass

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@visit can be called with Pass but the decorated function parameter annotations do not include this type",
        ):
            TestVisitor()

    def test_invalid_transformer_visit_superclass(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @visit(m.SimpleString() | m.Pass())
            def _string_visit(self, node: cst.BaseExpression) -> None:
                pass

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@visit can be called with Pass but the decorated function parameter annotations do not include this type",
        ):
            TestVisitor()

    def test_invalid_collector_leave_union(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @leave(m.SimpleString() | m.Name())
            def _string_leave(self, node: cst.SimpleString) -> None:
                pass

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave can be called with Name but the decorated function parameter annotations do not include this type",
        ):
            TestVisitor()

    def test_invalid_transformer_leave_union(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.SimpleString() | m.Name())
            def _string_leave(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> cst.BaseExpression:
                return updated_node

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave can be called with Name but the decorated function parameter annotations do not include this type",
        ):
            TestVisitor()

    def test_invalid_collector_leave_superclass(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @leave(m.SimpleString() | m.Pass())
            def _string_leave(self, node: cst.BaseExpression) -> None:
                pass

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave can be called with Pass but the decorated function parameter annotations do not include this type",
        ):
            TestVisitor()

    def test_invalid_transformer_leave_superclass(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.SimpleString() | m.Pass())
            def _string_leave(
                self,
                original_node: cst.BaseExpression,
                updated_node: cst.BaseExpression,
            ) -> cst.BaseExpression:
                return updated_node

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave can be called with Pass but the decorated function parameter annotations do not include this type",
        ):
            TestVisitor()

    def test_bad_visit_collecter_decorator(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @visit(m.SimpleString())
            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                pass

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@visit should not decorate functions that are concrete visit or leave methods",
        ):
            TestVisitor()

    def test_bad_leave_collecter_decorator(self) -> None:
        class TestVisitor(MatcherDecoratableVisitor):
            @leave(m.SimpleString())
            def leave_SimpleString(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> None:
                pass

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave should not decorate functions that are concrete visit or leave methods",
        ):
            TestVisitor()

    def test_bad_visit_transform_decorator(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @visit(m.SimpleString())
            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                pass

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@visit should not decorate functions that are concrete visit or leave methods",
        ):
            TestVisitor()

    def test_bad_leave_transform_decorator(self) -> None:
        class TestVisitor(MatcherDecoratableTransformer):
            @leave(m.SimpleString())
            def leave_SimpleString(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> cst.SimpleString:
                return updated_node

        # Instantiating this class should raise a runtime error
        with self.assertRaisesRegex(
            MatchDecoratorMismatch,
            "@leave should not decorate functions that are concrete visit or leave methods",
        ):
            TestVisitor()

    def test_pickleable_exception(self) -> None:
        original = MatchDecoratorMismatch("func", "message")
        serialized = pickle.dumps(original)
        unserialized = pickle.loads(serialized)
        self.assertEqual(original.message, unserialized.message)
        self.assertEqual(original.func, unserialized.func)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Callable, Optional

import libcst as cst
from libcst import parse_statement
from libcst._nodes.tests.base import CSTNodeTest, parse_statement_as
from libcst.helpers import ensure_type
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class YieldConstructionTest(CSTNodeTest):
    @data_provider(
        (
            # Simple yield
            (cst.Yield(), "yield"),
            # yield expression
            (cst.Yield(cst.Name("a")), "yield a"),
            # yield from expression
            (cst.Yield(cst.From(cst.Call(cst.Name("a")))), "yield from a()"),
            # Parenthesizing tests
            (
                cst.Yield(
                    lpar=(cst.LeftParen(),),
                    value=cst.Integer("5"),
                    rpar=(cst.RightParen(),),
                ),
                "(yield 5)",
            ),
            # Whitespace oddities tests
            (
                cst.Yield(
                    cst.Name("a", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)),
                    whitespace_after_yield=cst.SimpleWhitespace(""),
                ),
                "yield(a)",
                CodeRange((1, 0), (1, 8)),
            ),
            (
                cst.Yield(
                    cst.From(
                        cst.Call(
                            cst.Name("a"),
                            lpar=(cst.LeftParen(),),
                            rpar=(cst.RightParen(),),
                        ),
                        whitespace_after_from=cst.SimpleWhitespace(""),
                    )
                ),
                "yield from(a())",
            ),
            # Whitespace rendering/parsing tests
            (
                cst.Yield(
                    lpar=(cst.LeftParen(whitespace_after=cst.SimpleWhitespace(" ")),),
                    value=cst.Integer("5"),
                    whitespace_after_yield=cst.SimpleWhitespace("  "),
                    rpar=(cst.RightParen(whitespace_before=cst.SimpleWhitespace(" ")),),
                ),
                "( yield  5 )",
            ),
            (
                cst.Yield(
                    lpar=(cst.LeftParen(whitespace_after=cst.SimpleWhitespace(" ")),),
                    value=cst.From(
                        cst.Call(cst.Name("bla")),
                        whitespace_after_from=cst.SimpleWhitespace("  "),
                    ),
                    whitespace_after_yield=cst.SimpleWhitespace("  "),
                    rpar=(cst.RightParen(whitespace_before=cst.SimpleWhitespace(" ")),),
                ),
                "( yield  from  bla() )",
                CodeRange((1, 2), (1, 20)),
            ),
            # From expression position tests
            (
                cst.From(
                    cst.Integer("5"), whitespace_after_from=cst.SimpleWhitespace(" ")
                ),
                "from 5",
                CodeRange((1, 0), (1, 6)),
            ),
        )
    )
    def test_valid(
        self, node: cst.CSTNode, code: str, position: Optional[CodeRange] = None
    ) -> None:
        self.validate_node(node, code, expected_position=position)

    @data_provider(
        (
            # Paren validation
            (
                lambda: cst.Yield(lpar=(cst.LeftParen(),)),
                "left paren without right paren",
            ),
            (
                lambda: cst.Yield(rpar=(cst.RightParen(),)),
                "right paren without left paren",
            ),
            # Make sure we have adequate space after yield
            (
                lambda: cst.Yield(
                    cst.Name("a"), whitespace_after_yield=cst.SimpleWhitespace("")
                ),
                "Must have at least one space after 'yield' keyword",
            ),
            (
                lambda: cst.Yield(
                    cst.From(cst.Call(cst.Name("a"))),
                    whitespace_after_yield=cst.SimpleWhitespace(""),
                ),
                "Must have at least one space after 'yield' keyword",
            ),
            # MAke sure we have adequate space after from
            (
                lambda: cst.Yield(
                    cst.From(
                        cst.Call(cst.Name("a")),
                        whitespace_after_from=cst.SimpleWhitespace(""),
                    )
                ),
                "Must have at least one space after 'from' keyword",
            ),
        )
    )
    def test_invalid(
        self, get_node: Callable[[], cst.CSTNode], expected_re: str
    ) -> None:
        self.assert_invalid(get_node, expected_re)


class YieldParsingTest(CSTNodeTest):
    @data_provider(
        (
            # Simple yield
            (cst.Yield(), "yield"),
            # yield expression
            (
                cst.Yield(
                    cst.Name("a"), whitespace_after_yield=cst.SimpleWhitespace(" ")
                ),
                "yield a",
            ),
            # yield from expression
            (
                cst.Yield(
                    cst.From(
                        cst.Call(cst.Name("a")),
                        whitespace_after_from=cst.SimpleWhitespace(" "),
                    ),
                    whitespace_after_yield=cst.SimpleWhitespace(" "),
                ),
                "yield from a()",
            ),
            # Parenthesizing tests
            (
                cst.Yield(
                    lpar=(cst.LeftParen(),),
                    whitespace_after_yield=cst.SimpleWhitespace(" "),
                    value=cst.Integer("5"),
                    rpar=(cst.RightParen(),),
                ),
                "(yield 5)",
            ),
            # Whitespace oddities tests
            (
                cst.Yield(
                    cst.Name("a", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)),
                    whitespace_after_yield=cst.SimpleWhitespace(""),
                ),
                "yield(a)",
            ),
            (
                cst.Yield(
                    cst.From(
                        cst.Call(
                            cst.Name("a"),
                            lpar=(cst.LeftParen(),),
                            rpar=(cst.RightParen(),),
                        ),
                        whitespace_after_from=cst.SimpleWhitespace(""),
                    ),
                    whitespace_after_yield=cst.SimpleWhitespace(" "),
                ),
                "yield from(a())",
            ),
            # Whitespace rendering/parsing tests
            (
                cst.Yield(
                    lpar=(cst.LeftParen(whitespace_after=cst.SimpleWhitespace(" ")),),
                    value=cst.Integer("5"),
                    whitespace_after_yield=cst.SimpleWhitespace("  "),
                    rpar=(cst.RightParen(whitespace_before=cst.SimpleWhitespace(" ")),),
                ),
                "( yield  5 )",
            ),
            (
                cst.Yield(
                    lpar=(cst.LeftParen(whitespace_after=cst.SimpleWhitespace(" ")),),
                    value=cst.From(
                        cst.Call(cst.Name("bla")),
                        whitespace_after_from=cst.SimpleWhitespace("  "),
                    ),
                    whitespace_after_yield=cst.SimpleWhitespace("  "),
                    rpar=(cst.RightParen(whitespace_before=cst.SimpleWhitespace(" ")),),
                ),
                "( yield  from  bla() )",
            ),
        )
    )
    def test_valid(
        self, node: cst.CSTNode, code: str, position: Optional[CodeRange] = None
    ) -> None:
        self.validate_node(
            node,
            code,
            lambda code: ensure_type(
                ensure_type(parse_statement(code), cst.SimpleStatementLine).body[0],
                cst.Expr,
            ).value,
        )

    @data_provider(
        (
            {
                "code": "yield from x",
                "parser": parse_statement_as(python_version="3.3"),
                "expect_success": True,
            },
            {
                "code": "yield from x",
                "parser": parse_statement_as(python_version="3.1"),
                "expect_success": False,
            },
        )
    )
    def test_versions(self, **kwargs: Any) -> None:
        self.assert_parses(**kwargs)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from typing import Type

from libcst.codemod import Codemod, CodemodTest
from libcst.codemod.commands.unnecessary_format_string import UnnecessaryFormatString


class TestUnnecessaryFormatString(CodemodTest):
    TRANSFORM: Type[Codemod] = UnnecessaryFormatString

    def test_replace(self) -> None:
        before = r"""
            good: str = "good"
            good: str = f"with_arg{arg}"
            good = "good{arg1}".format(1234)
            good = "good".format()
            good = "good" % {}
            good = "good" % ()
            good = rf"good\d+{bar}"
            good = f"wow i don't have args but don't mess my braces {{ up }}"

            bad: str = f"bad" + "bad"
            bad: str = f'bad'
            bad: str = rf'bad\d+'
        """
        after = r"""
            good: str = "good"
            good: str = f"with_arg{arg}"
            good = "good{arg1}".format(1234)
            good = "good".format()
            good = "good" % {}
            good = "good" % ()
            good = rf"good\d+{bar}"
            good = f"wow i don't have args but don't mess my braces {{ up }}"

            bad: str = "bad" + "bad"
            bad: str = 'bad'
            bad: str = r'bad\d+'
        """
        self.assertCodemod(before, after)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from ast import literal_eval
from textwrap import dedent
from typing import List, Set

import libcst as cst
import libcst.matchers as m
from libcst.matchers import (
    MatcherDecoratableTransformer,
    MatcherDecoratableVisitor,
    call_if_inside,
    call_if_not_inside,
    leave,
    visit,
)
from libcst.testing.utils import UnitTest


def fixture(code: str) -> cst.Module:
    return cst.parse_module(dedent(code))


class MatchersGatingDecoratorsTest(UnitTest):
    def test_call_if_inside_transform_simple(self) -> None:
        # Set up a simple visitor with a call_if_inside decorator.
        class TestVisitor(MatcherDecoratableTransformer):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []
                self.leaves: List[str] = []

            @call_if_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                self.visits.append(node.value)

            @call_if_inside(m.FunctionDef())
            def leave_SimpleString(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> cst.SimpleString:
                self.leaves.append(updated_node.value)
                return updated_node

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ['"baz"'])
        self.assertEqual(visitor.leaves, ['"baz"', '"foobar"'])

    def test_call_if_inside_verify_original_transform(self) -> None:
        # Set up a simple visitor with a call_if_inside decorator.
        class TestVisitor(MatcherDecoratableTransformer):
            def __init__(self) -> None:
                super().__init__()
                self.func_visits: List[str] = []
                self.str_visits: List[str] = []

            @call_if_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                self.str_visits.append(node.value)

            def visit_FunctionDef(self, node: cst.FunctionDef) -> None:
                self.func_visits.append(node.name.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.func_visits, ["foo", "bar"])
        self.assertEqual(visitor.str_visits, ['"baz"'])

    def test_call_if_inside_collect_simple(self) -> None:
        # Set up a simple visitor with a call_if_inside decorator.
        class TestVisitor(MatcherDecoratableVisitor):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []
                self.leaves: List[str] = []

            @call_if_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                self.visits.append(node.value)

            @call_if_inside(m.FunctionDef())
            def leave_SimpleString(self, original_node: cst.SimpleString) -> None:
                self.leaves.append(original_node.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ['"baz"'])
        self.assertEqual(visitor.leaves, ['"baz"', '"foobar"'])

    def test_call_if_inside_verify_original_collect(self) -> None:
        # Set up a simple visitor with a call_if_inside decorator.
        class TestVisitor(MatcherDecoratableVisitor):
            def __init__(self) -> None:
                super().__init__()
                self.func_visits: List[str] = []
                self.str_visits: List[str] = []

            @call_if_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                self.str_visits.append(node.value)

            def visit_FunctionDef(self, node: cst.FunctionDef) -> None:
                self.func_visits.append(node.name.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.func_visits, ["foo", "bar"])
        self.assertEqual(visitor.str_visits, ['"baz"'])

    def test_multiple_visitors_collect(self) -> None:
        # Set up a simple visitor with multiple visit decorators.
        class TestVisitor(MatcherDecoratableVisitor):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []

            @call_if_inside(m.ClassDef(m.Name("A")))
            @call_if_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                self.visits.append(node.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            def foo() -> None:
                return "foo"

            class A:
                def foo(self) -> None:
                    return "baz"
            """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ['"baz"'])

    def test_multiple_visitors_transform(self) -> None:
        # Set up a simple visitor with multiple visit decorators.
        class TestVisitor(MatcherDecoratableTransformer):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []

            @call_if_inside(m.ClassDef(m.Name("A")))
            @call_if_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                self.visits.append(node.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            def foo() -> None:
                return "foo"

            class A:
                def foo(self) -> None:
                    return "baz"
            """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ['"baz"'])

    def test_call_if_not_inside_transform_simple(self) -> None:
        # Set up a simple visitor with a call_if_inside decorator.
        class TestVisitor(MatcherDecoratableTransformer):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []
                self.leaves: List[str] = []

            @call_if_not_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                self.visits.append(node.value)

            @call_if_not_inside(m.FunctionDef())
            def leave_SimpleString(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> cst.SimpleString:
                self.leaves.append(updated_node.value)
                return updated_node

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ['"foo"', '"bar"', '"foobar"'])
        self.assertEqual(visitor.leaves, ['"foo"', '"bar"'])

    def test_visit_if_inot_inside_verify_original_transform(self) -> None:
        # Set up a simple visitor with a call_if_inside decorator.
        class TestVisitor(MatcherDecoratableTransformer):
            def __init__(self) -> None:
                super().__init__()
                self.func_visits: List[str] = []
                self.str_visits: List[str] = []

            @call_if_not_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                self.str_visits.append(node.value)

            def visit_FunctionDef(self, node: cst.FunctionDef) -> None:
                self.func_visits.append(node.name.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.func_visits, ["foo", "bar"])
        self.assertEqual(visitor.str_visits, ['"foo"', '"bar"', '"foobar"'])

    def test_call_if_not_inside_collect_simple(self) -> None:
        # Set up a simple visitor with a call_if_inside decorator.
        class TestVisitor(MatcherDecoratableVisitor):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []
                self.leaves: List[str] = []

            @call_if_not_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                self.visits.append(node.value)

            @call_if_not_inside(m.FunctionDef())
            def leave_SimpleString(self, original_node: cst.SimpleString) -> None:
                self.leaves.append(original_node.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ['"foo"', '"bar"', '"foobar"'])
        self.assertEqual(visitor.leaves, ['"foo"', '"bar"'])

    def test_visit_if_inot_inside_verify_original_collect(self) -> None:
        # Set up a simple visitor with a call_if_inside decorator.
        class TestVisitor(MatcherDecoratableVisitor):
            def __init__(self) -> None:
                super().__init__()
                self.func_visits: List[str] = []
                self.str_visits: List[str] = []

            @call_if_not_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                self.str_visits.append(node.value)

            def visit_FunctionDef(self, node: cst.FunctionDef) -> None:
                self.func_visits.append(node.name.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.func_visits, ["foo", "bar"])
        self.assertEqual(visitor.str_visits, ['"foo"', '"bar"', '"foobar"'])


class MatchersVisitLeaveDecoratorsTest(UnitTest):
    def test_visit_transform(self) -> None:
        # Set up a simple visitor with a visit and leave decorator.
        class TestVisitor(MatcherDecoratableTransformer):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []
                self.leaves: List[str] = []

            @visit(m.FunctionDef(m.Name("foo") | m.Name("bar")))
            def visit_function(self, node: cst.FunctionDef) -> None:
                self.visits.append(node.name.value)

            @leave(m.FunctionDef(m.Name("bar") | m.Name("baz")))
            def leave_function(
                self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
            ) -> cst.FunctionDef:
                self.leaves.append(updated_node.name.value)
                return updated_node

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"

            def baz() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ["foo", "bar"])
        self.assertEqual(visitor.leaves, ["bar", "baz"])

    def test_visit_collector(self) -> None:
        # Set up a simple visitor with a visit and leave decorator.
        class TestVisitor(MatcherDecoratableVisitor):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []
                self.leaves: List[str] = []

            @visit(m.FunctionDef(m.Name("foo") | m.Name("bar")))
            def visit_function(self, node: cst.FunctionDef) -> None:
                self.visits.append(node.name.value)

            @leave(m.FunctionDef(m.Name("bar") | m.Name("baz")))
            def leave_function(self, original_node: cst.FunctionDef) -> None:
                self.leaves.append(original_node.name.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"

            def baz() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ["foo", "bar"])
        self.assertEqual(visitor.leaves, ["bar", "baz"])

    def test_stacked_visit_transform(self) -> None:
        # Set up a simple visitor with a visit and leave decorator.
        class TestVisitor(MatcherDecoratableTransformer):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []
                self.leaves: List[str] = []

            @visit(m.FunctionDef(m.Name("foo")))
            @visit(m.FunctionDef(m.Name("bar")))
            def visit_function(self, node: cst.FunctionDef) -> None:
                self.visits.append(node.name.value)

            @leave(m.FunctionDef(m.Name("bar")))
            @leave(m.FunctionDef(m.Name("baz")))
            def leave_function(
                self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
            ) -> cst.FunctionDef:
                self.leaves.append(updated_node.name.value)
                return updated_node

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"

            def baz() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ["foo", "bar"])
        self.assertEqual(visitor.leaves, ["bar", "baz"])

    def test_stacked_visit_collector(self) -> None:
        # Set up a simple visitor with a visit and leave decorator.
        class TestVisitor(MatcherDecoratableVisitor):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []
                self.leaves: List[str] = []

            @visit(m.FunctionDef(m.Name("foo")))
            @visit(m.FunctionDef(m.Name("bar")))
            def visit_function(self, node: cst.FunctionDef) -> None:
                self.visits.append(node.name.value)

            @leave(m.FunctionDef(m.Name("bar")))
            @leave(m.FunctionDef(m.Name("baz")))
            def leave_function(self, original_node: cst.FunctionDef) -> None:
                self.leaves.append(original_node.name.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"

            def baz() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ["foo", "bar"])
        self.assertEqual(visitor.leaves, ["bar", "baz"])
        self.assertEqual(visitor.leaves, ["bar", "baz"])

    def test_duplicate_visit_transform(self) -> None:
        # Set up a simple visitor with a visit and leave decorator.
        class TestVisitor(MatcherDecoratableTransformer):
            def __init__(self) -> None:
                super().__init__()
                self.visits: Set[str] = set()
                self.leaves: Set[str] = set()

            @visit(m.FunctionDef(m.Name("foo")))
            def visit_function1(self, node: cst.FunctionDef) -> None:
                self.visits.add(node.name.value + "1")

            @visit(m.FunctionDef(m.Name("foo")))
            def visit_function2(self, node: cst.FunctionDef) -> None:
                self.visits.add(node.name.value + "2")

            @leave(m.FunctionDef(m.Name("bar")))
            def leave_function1(
                self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
            ) -> cst.FunctionDef:
                self.leaves.add(updated_node.name.value + "1")
                return updated_node

            @leave(m.FunctionDef(m.Name("bar")))
            def leave_function2(
                self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
            ) -> cst.FunctionDef:
                self.leaves.add(updated_node.name.value + "2")
                return updated_node

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"

            def baz() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, {"foo1", "foo2"})
        self.assertEqual(visitor.leaves, {"bar1", "bar2"})

    def test_duplicate_visit_collector(self) -> None:
        # Set up a simple visitor with a visit and leave decorator.
        class TestVisitor(MatcherDecoratableVisitor):
            def __init__(self) -> None:
                super().__init__()
                self.visits: Set[str] = set()
                self.leaves: Set[str] = set()

            @visit(m.FunctionDef(m.Name("foo")))
            def visit_function1(self, node: cst.FunctionDef) -> None:
                self.visits.add(node.name.value + "1")

            @visit(m.FunctionDef(m.Name("foo")))
            def visit_function2(self, node: cst.FunctionDef) -> None:
                self.visits.add(node.name.value + "2")

            @leave(m.FunctionDef(m.Name("bar")))
            def leave_function1(self, original_node: cst.FunctionDef) -> None:
                self.leaves.add(original_node.name.value + "1")

            @leave(m.FunctionDef(m.Name("bar")))
            def leave_function2(self, original_node: cst.FunctionDef) -> None:
                self.leaves.add(original_node.name.value + "2")

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"

            def baz() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, {"foo1", "foo2"})
        self.assertEqual(visitor.leaves, {"bar1", "bar2"})

    def test_gated_visit_transform(self) -> None:
        # Set up a simple visitor with a visit and leave decorator.
        class TestVisitor(MatcherDecoratableTransformer):
            def __init__(self) -> None:
                super().__init__()
                self.visits: Set[str] = set()
                self.leaves: Set[str] = set()

            @call_if_inside(m.FunctionDef(m.Name("foo")))
            @visit(m.SimpleString())
            def visit_string1(self, node: cst.SimpleString) -> None:
                self.visits.add(literal_eval(node.value) + "1")

            @call_if_not_inside(m.FunctionDef(m.Name("bar")))
            @visit(m.SimpleString())
            def visit_string2(self, node: cst.SimpleString) -> None:
                self.visits.add(literal_eval(node.value) + "2")

            @call_if_inside(m.FunctionDef(m.Name("baz")))
            @leave(m.SimpleString())
            def leave_string1(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> cst.SimpleString:
                self.leaves.add(literal_eval(updated_node.value) + "1")
                return updated_node

            @call_if_not_inside(m.FunctionDef(m.Name("foo")))
            @leave(m.SimpleString())
            def leave_string2(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> cst.SimpleString:
                self.leaves.add(literal_eval(updated_node.value) + "2")
                return updated_node

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"

            def baz() -> None:
                return "foobarbaz"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, {"baz1", "foo2", "bar2", "baz2", "foobarbaz2"})
        self.assertEqual(
            visitor.leaves, {"foobarbaz1", "foo2", "bar2", "foobar2", "foobarbaz2"}
        )

    def test_gated_visit_collect(self) -> None:
        # Set up a simple visitor with a visit and leave decorator.
        class TestVisitor(MatcherDecoratableVisitor):
            def __init__(self) -> None:
                super().__init__()
                self.visits: Set[str] = set()
                self.leaves: Set[str] = set()

            @call_if_inside(m.FunctionDef(m.Name("foo")))
            @visit(m.SimpleString())
            def visit_string1(self, node: cst.SimpleString) -> None:
                self.visits.add(literal_eval(node.value) + "1")

            @call_if_not_inside(m.FunctionDef(m.Name("bar")))
            @visit(m.SimpleString())
            def visit_string2(self, node: cst.SimpleString) -> None:
                self.visits.add(literal_eval(node.value) + "2")

            @call_if_inside(m.FunctionDef(m.Name("baz")))
            @leave(m.SimpleString())
            def leave_string1(self, original_node: cst.SimpleString) -> None:
                self.leaves.add(literal_eval(original_node.value) + "1")

            @call_if_not_inside(m.FunctionDef(m.Name("foo")))
            @leave(m.SimpleString())
            def leave_string2(self, original_node: cst.SimpleString) -> None:
                self.leaves.add(literal_eval(original_node.value) + "2")

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"

            def baz() -> None:
                return "foobarbaz"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, {"baz1", "foo2", "bar2", "baz2", "foobarbaz2"})
        self.assertEqual(
            visitor.leaves, {"foobarbaz1", "foo2", "bar2", "foobar2", "foobarbaz2"}
        )

    def test_transform_order(self) -> None:
        # Set up a simple visitor with a visit and leave decorator.
        class TestVisitor(MatcherDecoratableTransformer):
            @call_if_inside(m.FunctionDef(m.Name("bar")))
            @leave(m.SimpleString())
            def leave_string1(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> cst.SimpleString:
                return updated_node.with_changes(
                    value=f'"prefix{literal_eval(updated_node.value)}"'
                )

            @call_if_inside(m.FunctionDef(m.Name("bar")))
            @leave(m.SimpleString())
            def leave_string2(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> cst.SimpleString:
                return updated_node.with_changes(
                    value=f'"{literal_eval(updated_node.value)}suffix"'
                )

            @call_if_inside(m.FunctionDef(m.Name("bar")))
            def leave_SimpleString(
                self, original_node: cst.SimpleString, updated_node: cst.SimpleString
            ) -> cst.SimpleString:
                return updated_node.with_changes(
                    value=f'"{"".join(reversed(literal_eval(updated_node.value)))}"'
                )

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"

            def baz() -> None:
                return "foobarbaz"
            """
        )
        visitor = TestVisitor()
        actual = module.visit(visitor)
        expected = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "prefixraboofsuffix"

            def baz() -> None:
                return "foobarbaz"
            """
        )
        self.assertTrue(expected.deep_equals(actual))

    def test_call_if_inside_visitor_attribute(self) -> None:
        # Set up a simple visitor with a call_if_inside decorator.
        class TestVisitor(MatcherDecoratableVisitor):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []
                self.leaves: List[str] = []

            @call_if_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString_lpar(self, node: cst.SimpleString) -> None:
                self.visits.append(node.value)

            @call_if_inside(m.FunctionDef())
            def leave_SimpleString_lpar(self, node: cst.SimpleString) -> None:
                self.leaves.append(node.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ['"baz"'])
        self.assertEqual(visitor.leaves, ['"baz"', '"foobar"'])

    def test_call_if_inside_transform_attribute(self) -> None:
        # Set up a simple visitor with a call_if_inside decorator.
        class TestVisitor(MatcherDecoratableTransformer):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []
                self.leaves: List[str] = []

            @call_if_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString_lpar(self, node: cst.SimpleString) -> None:
                self.visits.append(node.value)

            @call_if_inside(m.FunctionDef())
            def leave_SimpleString_lpar(self, node: cst.SimpleString) -> None:
                self.leaves.append(node.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ['"baz"'])
        self.assertEqual(visitor.leaves, ['"baz"', '"foobar"'])

    def test_call_if_not_inside_visitor_attribute(self) -> None:
        # Set up a simple visitor with a call_if_inside decorator.
        class TestVisitor(MatcherDecoratableVisitor):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []
                self.leaves: List[str] = []

            @call_if_not_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString_lpar(self, node: cst.SimpleString) -> None:
                self.visits.append(node.value)

            @call_if_not_inside(m.FunctionDef())
            def leave_SimpleString_lpar(self, node: cst.SimpleString) -> None:
                self.leaves.append(node.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ['"foo"', '"bar"', '"foobar"'])
        self.assertEqual(visitor.leaves, ['"foo"', '"bar"'])

    def test_call_if_not_inside_transform_attribute(self) -> None:
        # Set up a simple visitor with a call_if_inside decorator.
        class TestVisitor(MatcherDecoratableTransformer):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []
                self.leaves: List[str] = []

            @call_if_not_inside(m.FunctionDef(m.Name("foo")))
            def visit_SimpleString_lpar(self, node: cst.SimpleString) -> None:
                self.visits.append(node.value)

            @call_if_not_inside(m.FunctionDef())
            def leave_SimpleString_lpar(self, node: cst.SimpleString) -> None:
                self.leaves.append(node.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ['"foo"', '"bar"', '"foobar"'])
        self.assertEqual(visitor.leaves, ['"foo"', '"bar"'])

    def test_init_with_unhashable_types(self) -> None:
        # Set up a simple visitor with a call_if_inside decorator.
        class TestVisitor(MatcherDecoratableTransformer):
            def __init__(self) -> None:
                super().__init__()
                self.visits: List[str] = []

            @call_if_inside(
                m.FunctionDef(m.Name("foo"), params=m.Parameters([m.ZeroOrMore()]))
            )
            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                self.visits.append(node.value)

        # Parse a module and verify we visited correctly.
        module = fixture(
            """
            a = "foo"
            b = "bar"

            def foo() -> None:
                return "baz"

            def bar() -> None:
                return "foobar"
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        # We should have only visited a select number of nodes.
        self.assertEqual(visitor.visits, ['"baz"'])



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import libcst as cst
from libcst import parse_statement
from libcst._nodes.tests.base import CSTNodeTest
from libcst.helpers import ensure_type
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class ImportCreateTest(CSTNodeTest):
    @data_provider(
        (
            # Simple import statement
            {
                "node": cst.Import(names=(cst.ImportAlias(cst.Name("foo")),)),
                "code": "import foo",
            },
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar"))
                        ),
                    )
                ),
                "code": "import foo.bar",
            },
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar"))
                        ),
                    )
                ),
                "code": "import foo.bar",
            },
            # Comma-separated list of imports
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar"))
                        ),
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("baz"))
                        ),
                    )
                ),
                "code": "import foo.bar, foo.baz",
                "expected_position": CodeRange((1, 0), (1, 23)),
            },
            # Import with an alias
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar")),
                            asname=cst.AsName(cst.Name("baz")),
                        ),
                    )
                ),
                "code": "import foo.bar as baz",
            },
            # Import with an alias, comma separated
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar")),
                            asname=cst.AsName(cst.Name("baz")),
                        ),
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("baz")),
                            asname=cst.AsName(cst.Name("bar")),
                        ),
                    )
                ),
                "code": "import foo.bar as baz, foo.baz as bar",
            },
            # Combine for fun and profit
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar")),
                            asname=cst.AsName(cst.Name("baz")),
                        ),
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("insta"), cst.Name("gram"))
                        ),
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("baz"))
                        ),
                        cst.ImportAlias(
                            cst.Name("unittest"), asname=cst.AsName(cst.Name("ut"))
                        ),
                    )
                ),
                "code": "import foo.bar as baz, insta.gram, foo.baz, unittest as ut",
            },
            # Verify whitespace works everywhere.
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(
                                cst.Name("foo"),
                                cst.Name("bar"),
                                dot=cst.Dot(
                                    whitespace_before=cst.SimpleWhitespace(" "),
                                    whitespace_after=cst.SimpleWhitespace(" "),
                                ),
                            ),
                            asname=cst.AsName(
                                cst.Name("baz"),
                                whitespace_before_as=cst.SimpleWhitespace("  "),
                                whitespace_after_as=cst.SimpleWhitespace("  "),
                            ),
                            comma=cst.Comma(
                                whitespace_before=cst.SimpleWhitespace(" "),
                                whitespace_after=cst.SimpleWhitespace("  "),
                            ),
                        ),
                        cst.ImportAlias(
                            cst.Name("unittest"),
                            asname=cst.AsName(
                                cst.Name("ut"),
                                whitespace_before_as=cst.SimpleWhitespace("  "),
                                whitespace_after_as=cst.SimpleWhitespace("  "),
                            ),
                        ),
                    ),
                    whitespace_after_import=cst.SimpleWhitespace("  "),
                ),
                "code": "import  foo . bar  as  baz ,  unittest  as  ut",
                "expected_position": CodeRange((1, 0), (1, 46)),
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        (
            {
                "get_node": lambda: cst.Import(names=()),
                "expected_re": "at least one ImportAlias",
            },
            {
                "get_node": lambda: cst.Import(names=(cst.ImportAlias(cst.Name("")),)),
                "expected_re": "empty name identifier",
            },
            {
                "get_node": lambda: cst.Import(
                    names=(
                        cst.ImportAlias(cst.Attribute(cst.Name(""), cst.Name("bla"))),
                    )
                ),
                "expected_re": "empty name identifier",
            },
            {
                "get_node": lambda: cst.Import(
                    names=(
                        cst.ImportAlias(cst.Attribute(cst.Name("bla"), cst.Name(""))),
                    )
                ),
                "expected_re": "empty name identifier",
            },
            {
                "get_node": lambda: cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar")),
                            comma=cst.Comma(),
                        ),
                    )
                ),
                "expected_re": "trailing comma",
            },
            {
                "get_node": lambda: cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar"))
                        ),
                    ),
                    whitespace_after_import=cst.SimpleWhitespace(""),
                ),
                "expected_re": "at least one space",
            },
            {
                "get_node": lambda: cst.Import(
                    names=[
                        cst.ImportAlias(
                            name=cst.Attribute(
                                value=cst.Float(value="0."), attr=cst.Name(value="A")
                            )
                        )
                    ]
                ),
                "expected_re": "imported name must be a valid qualified name.",
            },
        )
    )
    def test_invalid(self, **kwargs: Any) -> None:
        self.assert_invalid(**kwargs)


class ImportParseTest(CSTNodeTest):
    @data_provider(
        (
            # Simple import statement
            {
                "node": cst.Import(names=(cst.ImportAlias(cst.Name("foo")),)),
                "code": "import foo",
            },
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar"))
                        ),
                    )
                ),
                "code": "import foo.bar",
            },
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar"))
                        ),
                    )
                ),
                "code": "import foo.bar",
            },
            # Comma-separated list of imports
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar")),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("baz"))
                        ),
                    )
                ),
                "code": "import foo.bar, foo.baz",
            },
            # Import with an alias
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar")),
                            asname=cst.AsName(cst.Name("baz")),
                        ),
                    )
                ),
                "code": "import foo.bar as baz",
            },
            # Import with an alias, comma separated
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar")),
                            asname=cst.AsName(cst.Name("baz")),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("baz")),
                            asname=cst.AsName(cst.Name("bar")),
                        ),
                    )
                ),
                "code": "import foo.bar as baz, foo.baz as bar",
            },
            # Combine for fun and profit
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("bar")),
                            asname=cst.AsName(cst.Name("baz")),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("insta"), cst.Name("gram")),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.ImportAlias(
                            cst.Attribute(cst.Name("foo"), cst.Name("baz")),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.ImportAlias(
                            cst.Name("unittest"), asname=cst.AsName(cst.Name("ut"))
                        ),
                    )
                ),
                "code": "import foo.bar as baz, insta.gram, foo.baz, unittest as ut",
            },
            # Verify whitespace works everywhere.
            {
                "node": cst.Import(
                    names=(
                        cst.ImportAlias(
                            cst.Attribute(
                                cst.Name("foo"),
                                cst.Name("bar"),
                                dot=cst.Dot(
                                    whitespace_before=cst.SimpleWhitespace(" "),
                                    whitespace_after=cst.SimpleWhitespace(" "),
                                ),
                            ),
                            asname=cst.AsName(
                                cst.Name("baz"),
                                whitespace_before_as=cst.SimpleWhitespace("  "),
                                whitespace_after_as=cst.SimpleWhitespace("  "),
                            ),
                            comma=cst.Comma(
                                whitespace_before=cst.SimpleWhitespace(" "),
                                whitespace_after=cst.SimpleWhitespace("  "),
                            ),
                        ),
                        cst.ImportAlias(
                            cst.Name("unittest"),
                            asname=cst.AsName(
                                cst.Name("ut"),
                                whitespace_before_as=cst.SimpleWhitespace("  "),
                                whitespace_after_as=cst.SimpleWhitespace("  "),
                            ),
                        ),
                    ),
                    whitespace_after_import=cst.SimpleWhitespace("  "),
                ),
                "code": "import  foo . bar  as  baz ,  unittest  as  ut",
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(
            parser=lambda code: ensure_type(
                parse_statement(code), cst.SimpleStatementLine
            ).body[0],
            **kwargs,
        )


class ImportFromCreateTest(CSTNodeTest):
    @data_provider(
        (
            # Simple from import statement
            {
                "node": cst.ImportFrom(
                    module=cst.Name("foo"), names=(cst.ImportAlias(cst.Name("bar")),)
                ),
                "code": "from foo import bar",
            },
            # From import statement with alias
            {
                "node": cst.ImportFrom(
                    module=cst.Name("foo"),
                    names=(
                        cst.ImportAlias(
                            cst.Name("bar"), asname=cst.AsName(cst.Name("baz"))
                        ),
                    ),
                ),
                "code": "from foo import bar as baz",
            },
            # Multiple imports
            {
                "node": cst.ImportFrom(
                    module=cst.Name("foo"),
                    names=(
                        cst.ImportAlias(cst.Name("bar")),
                        cst.ImportAlias(cst.Name("baz")),
                    ),
                ),
                "code": "from foo import bar, baz",
            },
            # Trailing comma
            {
                "node": cst.ImportFrom(
                    module=cst.Name("foo"),
                    names=(
                        cst.ImportAlias(cst.Name("bar"), comma=cst.Comma()),
                        cst.ImportAlias(cst.Name("baz"), comma=cst.Comma()),
                    ),
                ),
                "code": "from foo import bar,baz,",
                "expected_position": CodeRange((1, 0), (1, 23)),
            },
            # Star import statement
            {
                "node": cst.ImportFrom(module=cst.Name("foo"), names=cst.ImportStar()),
                "code": "from foo import *",
                "expected_position": CodeRange((1, 0), (1, 17)),
            },
            # Simple relative import statement
            {
                "node": cst.ImportFrom(
                    relative=(cst.Dot(),),
                    module=cst.Name("foo"),
                    names=(cst.ImportAlias(cst.Name("bar")),),
                ),
                "code": "from .foo import bar",
            },
            {
                "node": cst.ImportFrom(
                    relative=(cst.Dot(), cst.Dot()),
                    module=cst.Name("foo"),
                    names=(cst.ImportAlias(cst.Name("bar")),),
                ),
                "code": "from ..foo import bar",
            },
            # Relative only import
            {
                "node": cst.ImportFrom(
                    relative=(cst.Dot(), cst.Dot()),
                    module=None,
                    names=(cst.ImportAlias(cst.Name("bar")),),
                ),
                "code": "from .. import bar",
            },
            # Parenthesis
            {
                "node": cst.ImportFrom(
                    module=cst.Name("foo"),
                    lpar=cst.LeftParen(),
                    names=(
                        cst.ImportAlias(
                            cst.Name("bar"), asname=cst.AsName(cst.Name("baz"))
                        ),
                    ),
                    rpar=cst.RightParen(),
                ),
                "code": "from foo import (bar as baz)",
                "expected_position": CodeRange((1, 0), (1, 28)),
            },
            # Verify whitespace works everywhere.
            {
                "node": cst.ImportFrom(
                    relative=(
                        cst.Dot(
                            whitespace_before=cst.SimpleWhitespace(" "),
                            whitespace_after=cst.SimpleWhitespace(" "),
                        ),
                        cst.Dot(
                            whitespace_before=cst.SimpleWhitespace(" "),
                            whitespace_after=cst.SimpleWhitespace(" "),
                        ),
                    ),
                    module=cst.Name("foo"),
                    lpar=cst.LeftParen(whitespace_after=cst.SimpleWhitespace(" ")),
                    names=(
                        cst.ImportAlias(
                            cst.Name("bar"),
                            asname=cst.AsName(
                                cst.Name("baz"),
                                whitespace_before_as=cst.SimpleWhitespace("  "),
                                whitespace_after_as=cst.SimpleWhitespace("  "),
                            ),
                            comma=cst.Comma(
                                whitespace_before=cst.SimpleWhitespace(" "),
                                whitespace_after=cst.SimpleWhitespace("  "),
                            ),
                        ),
                        cst.ImportAlias(
                            cst.Name("unittest"),
                            asname=cst.AsName(
                                cst.Name("ut"),
                                whitespace_before_as=cst.SimpleWhitespace("  "),
                                whitespace_after_as=cst.SimpleWhitespace("  "),
                            ),
                        ),
                    ),
                    rpar=cst.RightParen(whitespace_before=cst.SimpleWhitespace(" ")),
                    whitespace_after_from=cst.SimpleWhitespace("  "),
                    whitespace_before_import=cst.SimpleWhitespace("  "),
                    whitespace_after_import=cst.SimpleWhitespace("  "),
                ),
                "code": "from   .  . foo  import  ( bar  as  baz ,  unittest  as  ut )",
                "expected_position": CodeRange((1, 0), (1, 61)),
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        (
            {
                "get_node": lambda: cst.ImportFrom(
                    module=None, names=(cst.ImportAlias(cst.Name("bar")),)
                ),
                "expected_re": "Must have a module specified",
            },
            {
                "get_node": lambda: cst.ImportFrom(module=cst.Name("foo"), names=()),
                "expected_re": "at least one ImportAlias",
            },
            {
                "get_node": lambda: cst.ImportFrom(
                    module=cst.Name("foo"),
                    names=(cst.ImportAlias(cst.Name("bar")),),
                    lpar=cst.LeftParen(),
                ),
                "expected_re": "left paren without right paren",
            },
            {
                "get_node": lambda: cst.ImportFrom(
                    module=cst.Name("foo"),
                    names=(cst.ImportAlias(cst.Name("bar")),),
                    rpar=cst.RightParen(),
                ),
                "expected_re": "right paren without left paren",
            },
            {
                "get_node": lambda: cst.ImportFrom(
                    module=cst.Name("foo"), names=cst.ImportStar(), lpar=cst.LeftParen()
                ),
                "expected_re": "cannot have parens",
            },
            {
                "get_node": lambda: cst.ImportFrom(
                    module=cst.Name("foo"),
                    names=cst.ImportStar(),
                    rpar=cst.RightParen(),
                ),
                "expected_re": "cannot have parens",
            },
            {
                "get_node": lambda: cst.ImportFrom(
                    module=cst.Name("foo"),
                    names=(cst.ImportAlias(cst.Name("bar")),),
                    whitespace_after_from=cst.SimpleWhitespace(""),
                ),
                "expected_re": "one space after from",
            },
            {
                "get_node": lambda: cst.ImportFrom(
                    module=cst.Name("foo"),
                    names=(cst.ImportAlias(cst.Name("bar")),),
                    whitespace_before_import=cst.SimpleWhitespace(""),
                ),
                "expected_re": "one space before import",
            },
            {
                "get_node": lambda: cst.ImportFrom(
                    module=cst.Name("foo"),
                    names=(cst.ImportAlias(cst.Name("bar")),),
                    whitespace_after_import=cst.SimpleWhitespace(""),
                ),
                "expected_re": "one space after import",
            },
        )
    )
    def test_invalid(self, **kwargs: Any) -> None:
        self.assert_invalid(**kwargs)


class ImportFromParseTest(CSTNodeTest):
    @data_provider(
        (
            # Simple from import statement
            {
                "node": cst.ImportFrom(
                    module=cst.Name("foo"), names=(cst.ImportAlias(cst.Name("bar")),)
                ),
                "code": "from foo import bar",
            },
            # From import statement with alias
            {
                "node": cst.ImportFrom(
                    module=cst.Name("foo"),
                    names=(
                        cst.ImportAlias(
                            cst.Name("bar"), asname=cst.AsName(cst.Name("baz"))
                        ),
                    ),
                ),
                "code": "from foo import bar as baz",
            },
            # Multiple imports
            {
                "node": cst.ImportFrom(
                    module=cst.Name("foo"),
                    names=(
                        cst.ImportAlias(
                            cst.Name("bar"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.ImportAlias(cst.Name("baz")),
                    ),
                ),
                "code": "from foo import bar, baz",
            },
            # Trailing comma
            {
                "node": cst.ImportFrom(
                    module=cst.Name("foo"),
                    names=(
                        cst.ImportAlias(
                            cst.Name("bar"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.ImportAlias(cst.Name("baz"), comma=cst.Comma()),
                    ),
                ),
                "code": "from foo import bar, baz,",
            },
            # Star import statement
            {
                "node": cst.ImportFrom(module=cst.Name("foo"), names=cst.ImportStar()),
                "code": "from foo import *",
            },
            # Simple relative import statement
            {
                "node": cst.ImportFrom(
                    relative=(cst.Dot(),),
                    module=cst.Name("foo"),
                    names=(cst.ImportAlias(cst.Name("bar")),),
                ),
                "code": "from .foo import bar",
            },
            {
                "node": cst.ImportFrom(
                    relative=(cst.Dot(), cst.Dot()),
                    module=cst.Name("foo"),
                    names=(cst.ImportAlias(cst.Name("bar")),),
                ),
                "code": "from ..foo import bar",
            },
            # Relative only import
            {
                "node": cst.ImportFrom(
                    relative=(cst.Dot(), cst.Dot()),
                    module=None,
                    names=(cst.ImportAlias(cst.Name("bar")),),
                ),
                "code": "from .. import bar",
            },
            # Parenthesis
            {
                "node": cst.ImportFrom(
                    module=cst.Name("foo"),
                    lpar=cst.LeftParen(),
                    names=(
                        cst.ImportAlias(
                            cst.Name("bar"), asname=cst.AsName(cst.Name("baz"))
                        ),
                    ),
                    rpar=cst.RightParen(),
                ),
                "code": "from foo import (bar as baz)",
            },
            # Verify whitespace works everywhere.
            {
                "node": cst.ImportFrom(
                    relative=(
                        cst.Dot(
                            whitespace_before=cst.SimpleWhitespace(""),
                            whitespace_after=cst.SimpleWhitespace("  "),
                        ),
                        cst.Dot(
                            whitespace_before=cst.SimpleWhitespace(""),
                            whitespace_after=cst.SimpleWhitespace(" "),
                        ),
                    ),
                    module=cst.Name("foo"),
                    lpar=cst.LeftParen(whitespace_after=cst.SimpleWhitespace(" ")),
                    names=(
                        cst.ImportAlias(
                            cst.Name("bar"),
                            asname=cst.AsName(
                                cst.Name("baz"),
                                whitespace_before_as=cst.SimpleWhitespace("  "),
                                whitespace_after_as=cst.SimpleWhitespace("  "),
                            ),
                            comma=cst.Comma(
                                whitespace_before=cst.SimpleWhitespace(" "),
                                whitespace_after=cst.SimpleWhitespace("  "),
                            ),
                        ),
                        cst.ImportAlias(
                            cst.Name("unittest"),
                            asname=cst.AsName(
                                cst.Name("ut"),
                                whitespace_before_as=cst.SimpleWhitespace("  "),
                                whitespace_after_as=cst.SimpleWhitespace("  "),
                            ),
                        ),
                    ),
                    rpar=cst.RightParen(whitespace_before=cst.SimpleWhitespace(" ")),
                    whitespace_after_from=cst.SimpleWhitespace("   "),
                    whitespace_before_import=cst.SimpleWhitespace("  "),
                    whitespace_after_import=cst.SimpleWhitespace("  "),
                ),
                "code": "from   .  . foo  import  ( bar  as  baz ,  unittest  as  ut )",
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(
            parser=lambda code: ensure_type(
                parse_statement(code), cst.SimpleStatementLine
            ).body[0],
            **kwargs,
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Used by visitors. This is hoisted into a separate module to avoid some circular
dependencies in the definition of CSTNode.
"""

from enum import Enum, auto


class RemovalSentinel(Enum):
    """
    A :attr:`RemovalSentinel.REMOVE` value should be returned by a
    :meth:`CSTTransformer.on_leave` method when we want to remove that child from its
    parent. As a convenience, this can be constructed by calling
    :func:`libcst.RemoveFromParent`.

    The parent node should make a best-effort to remove the child, but may raise an
    exception when removing the child doesn't make sense, or could change the semantics
    in an unexpected way. For example, a function definition with no name doesn't make
    sense, but removing one of the arguments is valid.

    In we can't automatically remove the child, the developer should instead remove the
    child by constructing a new parent in the parent's :meth:`~CSTTransformer.on_leave`
    call.

    We use this instead of ``None`` to force developers to be explicit about deletions.
    Because ``None`` is the default return value for a function with no return
    statement, it would be too easy to accidentally delete nodes from the tree by
    forgetting to return a value.
    """

    REMOVE = auto()


def RemoveFromParent() -> RemovalSentinel:
    """
    A convenience method for requesting that this node be removed by its parent.
    Use this in place of returning :class:`RemovalSentinel` directly.
    For example, to remove all arguments unconditionally::

        def leave_Arg(
            self, original_node: cst.Arg, updated_node: cst.Arg
        ) -> Union[cst.Arg, cst.RemovalSentinel]:
            return RemoveFromParent()
    """
    return RemovalSentinel.REMOVE
